# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04_predict.ipynb (unless otherwise specified).

__all__ = ['open_nc', 'crop', 'image2tiles', 'tiles2image', 'get_preds', 'predict_one', 'save_netcdf', 'inspect_netcdf',
           'load_netcdf', 'predict_time', 'predict_month', 'predict_nrt', 'split_mask']

# Cell
import scipy.io as sio
import sys
from tqdm import tqdm
import netCDF4
import pdb
import gc
import calendar
import scipy.ndimage as ndimage
from fastai.vision.all import *

from .core import *
from .geo import Region
from .models import BA_Net

# Cell
def open_nc(fn, slice_idx=None, *args, **kwargs):
    data = netCDF4.Dataset(fn, mode='r')
    if slice_idx is None:
        data = np.array([data[r][:] for r in ['Red', 'NIR', 'MIR', 'FRP']], dtype=np.float32)
    else:
        data = np.array([data[r][slice_idx[0]:slice_idx[1], slice_idx[2]:slice_idx[3]]
                         for r in ['Red', 'NIR', 'MIR', 'FRP']], dtype=np.float32)
    data[data == 65535] = 0
    data = data/100
    return data.astype(np.float16)

def crop(im, r, c, size=128):
    '''
    crop image into a square of size sz,
    '''
    sz = size
    out_sz = (sz, sz, im.shape[-1])
    rs,cs,hs = im.shape
    tile = np.zeros(out_sz, dtype=np.float16)
    if (r+sz > rs) and (c+sz > cs):
        tile[:rs-r, :cs-c, :] = im[r:, c:, :]
    elif (r+sz > rs):
        tile[:rs-r, :, :] = im[r:, c:c+sz, :]
    elif (c+sz > cs):
        tile[:, :cs-c, :] = im[r:r+sz ,c:, :]
    else:
        tile[...] = im[r:r+sz, c:c+sz, :]
    return tile

def image2tiles(x, step=100, size=128):
    tiles = []
    rr, cc, _ = x.shape
    for c in range(0, cc-1, step):
        for r in range(0, rr-1, step):
            img = crop(x, r, c, size=size)
            tiles.append(img)
    return np.array(tiles, dtype=np.float16)

def tiles2image(tiles, image_size, size=128, step=100):
    rr, cc, = image_size
    sz = size
    im = np.zeros(image_size, dtype=np.float16)
    indicator = np.zeros_like(im)
    k = 0
    for c in range(0, cc-1, step):
        for r in range(0, rr-1, step):
            if (r+sz > rr) and (c+sz > cc):
                im[r:, c:] += tiles[k][:rr-r, :cc-c]
                indicator[r:, c:] += 1
            elif (r+sz > rr):
                im[r:, c:c+sz] += tiles[k][:rr-r, :]
                indicator[r:, c:c+sz] += 1
            elif (c+sz > cc):
                im[r:r+sz ,c:] += tiles[k][:, :cc-c]
                indicator[r:r+sz, c:] += 1
            else:
                im[r:r+sz, c:c+sz] += tiles[k]
                indicator[r:r+sz, c:c+sz] += 1
            k += 1
    im /= indicator
    return im

def get_preds(tiles, model, weights=None, verbose=False):
    if weights is not None:
        model.load_state_dict(weights)
    mu = tensor([0.2349, 0.3548, 0.1128, 0.0016]).view(1,4,1,1,1)
    std = tensor([0.1879, 0.1660, 0.0547, 0.0776]).view(1,4,1,1,1)
    with torch.no_grad():
        data = []
        iterator = tqdm(tiles) if verbose else tiles
        for x in iterator:
            model.eval()
            if torch.cuda.is_available():
                model.cuda()
            x = (x[None]-mu)/std
            if torch.cuda.is_available():
                x = x.cuda()
            out = model(x).sigmoid().float()
            data.append(out.cpu().squeeze().numpy())
    return np.array(data, np.float16)

def predict_one(iop:InOutPath, times:list, weights_files:list, region:str, threshold=0.5,
                slice_idx=None, product='VIIRS750', verbose=False):
    fname = lambda t : iop.src/f'{product}{region}_{t.strftime("%Y%m%d")}.nc'
    files = [fname(t) for t in times]
    im_size = open_nc(files[0], slice_idx=slice_idx).shape[1:]
    tiles = []
    print('Loading data and generating tiles:')
    iterator = tqdm(files) if verbose else files
    for file in iterator:
        try:
            s = image2tiles(open_nc(file, slice_idx=slice_idx).transpose((1,2,0))).transpose((0,3,1,2))
        except:
            warn(f'No data for {file}')
            s = np.zeros_like(s)
        tiles.append(s)
    del s
    tiles = np.array(tiles, dtype=np.float16).transpose((1, 2, 0, 3, 4))
    tiles = torch.from_numpy(tiles).float()
    preds_ens = []
    for wf in weights_files:
        if torch.cuda.is_available():
            weights = torch.load(wf)
        else: weights = torch.load(wf, map_location=torch.device('cpu'))
        if 'model' in weights:
            weights = weights['model']
        print(f'Generating model predictions for {wf}:')
        preds = get_preds(tiles, model=BA_Net(4, 1, 64), weights=weights, verbose=verbose)
        preds = np.array([tiles2image(preds[:,i], im_size) for i in range(preds.shape[1])])
        preds_ens.append(preds)
    del tiles
    preds = np.array(preds_ens, dtype=np.float16).mean(0)
    return preds.astype(np.float16)

def save_netcdf(file, data:dict, region:Region, crop=None):
    mode = "r+" if Path(file).is_file() and crop is not None else "w"
    with netCDF4.Dataset(file, mode, format="NETCDF4") as rootgrp:
        if mode == "w":
            timeD = rootgrp.createDimension("time", None)
            latD = rootgrp.createDimension("lat", region.height)
            lonD = rootgrp.createDimension("lon", region.width)
        for k in data:
            if mode == "w":
                a = rootgrp.createVariable(k,"f4",("time","lat","lon",), zlib=True)
            else: a = rootgrp[k]
            if crop is None:
                a[:] = data[k]
            else:
                a[:,crop[0]:crop[1],crop[2]:crop[3]] = data[k]

def inspect_netcdf(file, var='preds', crop=None):
    if file.is_file():
        with netCDF4.Dataset(file, 'r') as ncdata:
            if crop is not None:
                is_missing = np.sum(ncdata[var][:,crop[0]:crop[1],crop[2]:crop[3]].mask)>0
            else:
                is_missing = np.sum(ncdata[var][:].mask)>0
    else: is_missing = True
    return is_missing

def load_netcdf(file, crop=None):
    with netCDF4.Dataset(file, 'r', format="NETCDF4") as ncdata:
        if crop is None:
            return ncdata['preds'][:].astype(np.float16)
        else:
            return ncdata['preds'][:, crop[0]:crop[1], crop[2]:crop[3]].astype(np.float16)

def predict_time(path:InOutPath, times, weight_files:list, region:Region,
                 threshold=0.05, save=True, max_size=2000, buffer=128,
                 product='VIIRS750', output='data', check_file=False, verbose=False):
    tstart, tend = times.min(), times.max()
    tstart = pd.Timestamp(f'{tstart.year}-{tstart.month}-01')
    tend = pd.Timestamp(f'{tend.year}-{tend.month}-01')
    ptimes = pd.date_range(tstart, tend, freq='MS')[1:-1]
    ptimes_eom = pd.date_range(tstart, tend, freq='M')[1:]
    si = [[max(0,j*max_size-buffer), (j+1)*max_size+buffer,
           max(0,i*max_size-buffer), (i+1)*max_size+buffer]
          for i in range(region.shape[1]//max_size+1) for j in range(region.shape[0]//max_size+1)]
    for time in progress_bar(ptimes):
        tstr = time.strftime('%Y%m')
        file_save = path.dst/f'{output}_{region.name}{tstr}.nc'
        if not file_save.is_file() or check_file:
            if verbose: print(f'Processing time {time}')
            time_start = pd.Timestamp((time - pd.Timedelta(days=15)).strftime('%Y-%m-15')) # Day 15, previous month
            times = pd.date_range(time_start, periods=64, freq='D')
            preds_all = []
            for i, split in enumerate(si):
                if verbose: print(f'Split {split}')
                if check_file:
                    if verbose: print(f'Checking for saved data in the existing netcdf file')
                    is_missing = inspect_netcdf(file_save, crop=split)
                else: is_missing = True
                if is_missing:
                    preds = predict_one(path, times, weight_files, region.name, slice_idx=split,
                                        product=product, verbose=verbose)
                    preds = preds[times.month == time.month]
                    save_netcdf(file_save, {'preds': preds}, region, crop=split)
                    del preds
                    gc.collect()
        else: warn(f'File {file_save} already exists, skipping to next month.')

    # Merging all months and computing confidence level and date of burning
    si = [[max(0,j*max_size-buffer), (j+1)*max_size+buffer,
           max(0,i*max_size-buffer), (i+1)*max_size+buffer]
           for i in range(region.shape[1]//max_size+1) for j in range(region.shape[0]//max_size+1)]
    if verbose: print('Merging monthly files')
    ylen = 366 if calendar.isleap(ptimes[0].year) else 365
    bd_all, ba_all = np.zeros(region.shape, dtype=np.float16), np.zeros(region.shape, dtype=np.float16)
    for i, split in enumerate(si):
        print(f'Split {split}')
        preds_all = []
        for time in progress_bar(ptimes):
            tstr = time.strftime('%Y%m')
            file = path.dst/f'{output}_{region.name}{tstr}.nc'
            data = load_netcdf(file, crop=split)
            preds_all.append(data)
        preds_all = np.concatenate(preds_all, axis=0)
        ba = preds_all.sum(0)
        ba[ba>1] = 1
        ba[ba<threshold] = np.nan
        bd = preds_all.argmax(0)
        bd = bd.astype(np.float16)
        bd[np.isnan(ba)] = np.nan
        ba_all[split[0]:split[1], split[2]:split[3]] = ba
        bd_all[split[0]:split[1], split[2]:split[3]] = bd
        del preds_all, ba, bd
        gc.collect()
    times = pd.date_range(ptimes[0], ptimes_eom[-1], freq='D')
    tstr_start = times[0].strftime('%Y')
    file_save = path.dst/f'{output}_{region.name}{tstr_start}.mat'
    sio.savemat(file_save, {'burned': ba_all, 'date': bd_all, 'times': times.astype(str).tolist()},
                do_compression=True)

def predict_month(iop, time, weight_files, region, threshold=0.5, save=True, slice_idx=None):
    time_start = pd.Timestamp((time - pd.Timedelta(days=30)).strftime('%Y-%m-15')) # Day 15, previous month
    times = pd.date_range(time_start, periods=64, freq='D')
    preds = predict_one(iop, times, weight_files, region, threshold=threshold, slice_idx=slice_idx)
    assert preds.shape[0] == len(times)
    preds = preds[times.month == time.month]
    ba = preds.sum(0)
    bd = preds.argmax(0)
    doy = np.asarray(pd.DatetimeIndex(times).dayofyear)
    bd = doy[bd].astype(float)
    bd[bd==doy[0]] = np.nan
    bd[ba<threshold] = np.nan
    ba[ba<threshold] = np.nan
    ba[ba>1] = 1
    if not save: return ba, bd
    tstr = time.strftime('%Y%m')
    sio.savemat(iop.dst/f'ba_{region}_{tstr}.mat', {'burned': ba, 'date': bd}, do_compression=True)

def predict_nrt(path:InOutPath, time, weight_files:list, region:Region,
                 threshold=0.1, save=True, max_size=2000, buffer=128,
                 product='VIIRS750', verbose=False):
    times = pd.date_range(time-pd.Timedelta(days=63), time, freq='D')
    si = [[max(0,j*max_size-buffer), (j+1)*max_size+buffer,
           max(0,i*max_size-buffer), (i+1)*max_size+buffer]
          for i in range(region.shape[1]//max_size+1) for j in range(region.shape[0]//max_size+1)]

    bas, bds = [], []
    for i, split in progress_bar(enumerate(si), total=len(si)):
        print(f'Split {split}')
        preds_all = predict_one(path, times, weight_files, region.name, slice_idx=split,
                                product=product, verbose=verbose)
        ba = preds_all.sum(0)
        ba[ba>1] = 1
        ba[ba<threshold] = np.nan
        bd = preds_all.argmax(0)
        bd = bd.astype(np.float16)
        bd[np.isnan(ba)] = np.nan
        bas.append(ba.astype(np.float16))
        bds.append(bd.astype(np.float16))
        gc.collect()
    ba_all = np.zeros(region.shape, dtype=np.float16)
    bd_all = np.zeros_like(ba_all)
    for i, split_idx in enumerate(si):
        ba_all[split_idx[0]:split_idx[1], split_idx[2]:split_idx[3]] = bas[i]
        bd_all[split_idx[0]:split_idx[1], split_idx[2]:split_idx[3]] = bds[i]
    if not save: return ba_all, bd_all
    tstr = time.strftime('%Y%m%d')
    sio.savemat(path.dst/f'ba_{tstr}.mat', {'burned': ba_all, 'date': bd_all, 'times': np.array(times).astype(str)}, do_compression=True)

def split_mask(mask, thr=0.5, thr_obj=1):
    labled, n_objs = ndimage.label(mask > thr)
    result = []
    for i in range(n_objs):
        obj = (labled == i + 1).astype(int)
        if (obj.sum() > thr_obj): result.append(obj)
    return result


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_core.ipynb (unless otherwise specified).

__all__ = ['filter_files', 'ls', 'hdf_attr_check', 'dict2json', 'monthlen', 'InOutPath', 'ProjectPath']

# Cell
import numpy as np
import json
from pathlib import Path
from nbdev.imports import test_eq

# Cell
def filter_files(files, include=[], exclude=[]):
    "Filter list of files using a list of strings to inculde and/or exclude"
    for incl in include:
        files = [f for f in files if incl in f.name]
    for excl in exclude:
        files = [f for f in files if excl not in f.name]
    return sorted(files)

def ls(x, recursive=False, include=[], exclude=[]):
    "List files in folder, if recursive is True also list subfolders"
    if not recursive:
        out = list(x.iterdir())
    else:
        out = [o for o in x.glob('**/*')]
    out = filter_files(out, include=include, exclude=exclude)
    return sorted(out)

Path.ls = ls

def hdf_attr_check(attr, hdf, default):
    "Check if attribute is in hdf_attr_dict and return default"
    return default if not hasattr(hdf, attr) else hdf.__getattr__(attr)

def dict2json(data:dict, file):
    "Writes json file from dict"
    with open(file, 'w') as f:
        f.write(json.dumps(data))

# Cell
def monthlen(year, month):
    "Gives lenght of the month"
    base = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    if (year % 4) == 0:
        if (year % 100) == 0:
            if (year % 400) == 0:
                base[1] += 1
        else:
            base[1] += 1
    return base[month-1]

# Cell
class InOutPath():
    """Keeps track of an input and a output path. Creates paths if they don't exist and mkdir=True"""
    def __init__(self, input_path:str, output_path:str, mkdir=True):
        if isinstance(input_path, str): input_path = Path(input_path)
        if isinstance(output_path, str): output_path = Path(output_path)
        self.input_path = input_path
        self.output_path = output_path
        if mkdir: self.mkdirs()

    @property
    def src(self):
        "Shortcut to input_path"
        return self.input_path

    @property
    def dst(self):
        "Shortcut to output_path"
        return self.output_path

    def mkdirs(self):
        self.input_path.mkdir(exist_ok=True, parents=True)
        self.output_path.mkdir(exist_ok=True, parents=True)

    def __truediv__(self, s):
        return InOutPath(self.src/s, self.dst/s)

    def __repr__(self):
        return '\n'.join([f'{i}: {o}' for i, o in self.__dict__.items()]) + '\n'

# Cell
class ProjectPath():
    def __init__(self, path:Path, config_dir='config', ladsweb_dir='ladsweb',
                 hotspots_dir='hotspots', dataset_dir='dataset', outputs_dir='outputs',
                 web_dir='web'):
        if isinstance(path, str): path = Path(path)
        self.path     = path
        self.config   = path/config_dir
        self.ladsweb  = path/ladsweb_dir
        self.hotspots = path/hotspots_dir
        self.dataset  = path/dataset_dir
        self.outputs  = path/outputs_dir
        self.web      = path/web_dir
        self.create_dirs([self.path, self.config, self.ladsweb, self.hotspots,
                          self.dataset, self.outputs, self.web])

    @staticmethod
    def create_dirs(paths_list):
        "Create directories in list."
        for o in paths_list: o.mkdir(exist_ok=True)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/06_cli.ipynb (unless otherwise specified).

__all__ = ['banet_viirs375_download', 'banet_viirs750_download', 'banet_create_dataset', 'banet_dataset2tiles',
           'banet_predict_monthly', 'banet_predict_times', 'banet_nrt_run']

# Cell
import calendar
import pandas as pd
from fastscript import call_parse, Param
import pdb
import os
import IPython
import pdb

from geoget.download import run_all

from .core import InOutPath, Path, ls, dict2json
from .data import *
from .geo import Region
from .predict import predict_month, predict_time
from .models import BA_Net
#from banet.train import train_model
from .nrt import ProjectPath, RunManager

# Cell
Path.ls = ls
_bands =  ['Reflectance_M5', 'Reflectance_M7', 'Reflectance_M10', 'Radiance_M12',
           'Radiance_M15', 'SolarZenithAngle', 'SatelliteZenithAngle']
_regions_path = '../data/regions'
_weight_files = ['models/banetv0.20-val2017-fold0.pth',
                 'models/banetv0.20-val2017-fold1.pth',
                 'models/banetv0.20-val2017-fold2.pth']
_weight_files = [os.path.expanduser('~/.banet/') + o for o in _weight_files]

# Cell
@call_parse
def banet_viirs375_download(region:Param("Region name", str),
    tstart:Param("Start of serach window yyyy-mm-dd HH:MM:SS", str),
    tend:Param("End of search windo yyyy-mm-dd HH:MM:SS", str),
    path_save:Param("Path to save the outputs of the request", str),
    regions_path:Param("Path for region json files", str)):
    Path(path_save).mkdir(exist_ok=True)
    region = Region.load(f'{regions_path}/R_{region}.json')
    viirs_downloader1 = VIIRS375_download(region, tstart, tend)
    viirs_downloader2 = VIIRS750_download(region, tstart, tend, bands=['SolarZenithAngle',
                      'SatelliteZenithAngle'])
    viirs_downloader_list1 = viirs_downloader1.split_times()
    viirs_downloader_list2 = viirs_downloader2.split_times()
    viirs_downloader_list = [*viirs_downloader_list1, *viirs_downloader_list2]
    print(f'Splitting request into {len(viirs_downloader_list)} orders.')
    run_all(viirs_downloader_list, path_save)

# Cell
@call_parse
def banet_viirs750_download(region:Param("Region name", str),
    tstart:Param("Start of serach window yyyy-mm-dd HH:MM:SS", str),
    tend:Param("End of search windo yyyy-mm-dd HH:MM:SS", str),
    path_save:Param("Path to save the outputs of the request", str),
    regions_path:Param("Path for region json files", str)):
    Path(path_save).mkdir(exist_ok=True)
    region = Region.load(f'{regions_path}/R_{region}.json')
    viirs_downloader = VIIRS750_download(region, tstart, tend)
    viirs_downloader_list = viirs_downloader.split_times()
    print(f'Splitting request into {len(viirs_downloader_list)} orders.')
    run_all(viirs_downloader_list, path_save)

# Cell
@call_parse
def banet_create_dataset(region:Param("Region name", str),
                   viirs_path:Param("Input path for VIIRS raw data", str),
                   fires_path:Param("Input path for Active Fires csv", str),
                   save_path:Param("Path to save outputs", str),
                   regions_path:Param("Path where region defenition files are stored", str),
                   mcd64_path:Param("Input path for MCD64 raw data", str)=None,
                   cci51_path:Param("Input path for FireCCI51 raw data", str)=None,
                   bands:Param("List of bands to use as inputs for VIIRS raw data", str)=_bands,
                   year:Param("Set to process a single year instead of all available", int)=None):

    paths = InOutPath(f'{viirs_path}', f'{save_path}')
    R = Region.load(f'{regions_path}/R_{region}.json')

    # VIIRS750
    print('\nCreating dataset for VIIRS750')
    viirs = Viirs750Dataset(paths, R, bands=bands)
    viirs.filter_times(year)
    merge_tiles = MergeTiles('SatelliteZenithAngle')
    mir_calc = MirCalc('SolarZenithAngle', 'Radiance_M12', 'Radiance_M15')
    rename = BandsRename(['Reflectance_M5', 'Reflectance_M7'], ['Red', 'NIR'])
    bfilter = BandsFilter(['Red', 'NIR', 'MIR'])
    act_fires = ActiveFiresLog(f'{fires_path}/hotspots{R.name}.csv')
    viirs.process_all(proc_funcs=[BandsAssertShape(), merge_tiles, mir_calc,
                                  rename, bfilter, act_fires])

    # MCD64A1C6
    if mcd64_path is not None:
        print('\nCreating dataset for MCD64A1C6')
        paths.input_path = Path(mcd64_path)
        mcd = MCD64Dataset(paths, R)
        mcd.match_times(viirs)
        mcd.process_all()

    # FireCCI51
    if cci51_path is not None:
        print('\nCreating dataset for FireCCI51')
        paths.input_path = Path(cci51_path)
        cci51 = FireCCI51Dataset(paths, R)
        cci51.match_times(viirs)
        cci51.process_all()

# Cell
@call_parse
def banet_dataset2tiles(region:Param("Region name", str),
                  input_path:Param("Input path for dataset", str),
                  output_path:Param("Output path for tiles dataset", str),
                  size:Param("Tiles size", int)=128,
                  step:Param("Step size of moving window to create tiles", int)=100,
                  year:Param("Set to process a single year instead of all available", int)=None):

    iop = InOutPath(input_path, output_path)
    r2t = Region2Tiles(iop, 'VIIRS750', 'MCD64A1C6', regions=[region],
                       bands=[['Red', 'NIR', 'MIR', 'FRP'], ['bafrac']],
                       size=size, step=step)
    if year is None: r2t.process_all()
    else: r2t.process_all(include=[f'_{year}'])

# Cell
@call_parse
def banet_predict_monthly(region:Param("Region name", str),
                    input_path:Param("Input path for dataset", str),
                    output_path:Param("Output path for tiles dataset", str),
                    year:Param("Set to process a single year instead of all available", int),
                    weight_files:Param("List of pth weight files", list)=_weight_files):

    iop = InOutPath(input_path, f'{output_path}')
    times = pd.DatetimeIndex([pd.Timestamp(o.stem.split('_')[-1])
                              for o in iop.src.ls(include=['.mat'])])
    times = times[times.year == year]
    tstart, tend = times.min(), times.max()
    month_start = (tstart + pd.Timedelta(days=31)).month
    for m in range(month_start, tend.month):
        print(f'Generating maps for {calendar.month_name[m]} {year}:')
        t = pd.Timestamp(f'{year}-{m}-01')
        predict_month(iop, t, weight_files, region)

# Cell
@call_parse
def banet_predict_times(region:Param("Region name", str),
                    tstart:Param("Start of search window yyyy-mm-01", str),
                    tend:Param("End of search window yyyy-mm-01", str),
                    input_path:Param("Input path for dataset", str),
                    output_path:Param("Output path for tiles dataset", str),
                    regions_path:Param("Path for region json files", str),
                    product:Param("Name of product (default VIIRS750)", str)="VIIRS750",
                    output:Param("Name of file to save results", str)="data",
                    weight_files:Param("List of pth weight files", list)=_weight_files):

    iop = InOutPath(input_path, f'{output_path}')

    tstart = pd.Timestamp(tstart) - pd.Timedelta(days=15)
    tstart = pd.Timestamp(f'{tstart.year}-{tstart.month}-01')
    tend = pd.Timestamp(tend) + pd.Timedelta(days=75)
    tend = pd.Timestamp(f'{tend.year}-{tend.month}-01') - pd.Timedelta(days=1)
    times = pd.date_range(tstart, tend, freq='D')

    R = Region.load(f'{regions_path}/R_{region}.json')
    predict_time(iop, times, weight_files, R, product=product, output=output)

# Cell
@call_parse
def banet_nrt_run(region:Param("Region name", str),
                  left:Param("Left limit of the bounding box.", float),
                  bottom:Param("Bottom limit of the bounding box.", float),
                  right:Param("Right limit of the bounding box.", float),
                  top:Param("Top limit of the bounding box.", float),
                  project_path:Param("Root directory of the project", str),
                  hotspots_region:Param("Hotspots region name", str),
                  product:Param("Product name", str)='VIIRS750',
                  area_epsg:Param("EPSG to compute burned area", int)=3763,
                  pixel_size:Param("Pixel size", float)=0.01,
                  time:Param("Day for the run", str, choices=["today", "yesterday"])="today",
                  threshold:Param("Threshold to apply to output of the model", float)=0.5,
                  skip_hotspots:Param("Skip download of ladsweb data", bool)=False,
                  skip_ladsweb:Param("Skip download of ladsweb data", bool)=False,
                  skip_preprocess:Param("Skip download of ladsweb data", bool)=False,
                  skip_getpreds:Param("Skip download of ladsweb data", bool)=False,
                  skip_postprocess:Param("Skip post processing", bool)=False,
                  single_fold:Param("Use only one model", bool)=False,
                  max_size:Param("Max size to split region on inference and avoid memory error", int)=2000):
    paths = ProjectPath(project_path)
    weight_files = ['banetv0.20-val2017-fold0.pth',
                    'banetv0.20-val2017-fold1.pth',
                    'banetv0.20-val2017-fold2.pth']
    if single_fold: weight_files = ['banetv0.20-val2017-fold0.pth']
    manager = RunManager(paths, region, time=time, product=product, max_size=max_size)
    R = {'name': region, 'bbox': [left, bottom, right, top], 'pixel_size': pixel_size}
    dict2json(R, paths.config/f'R_{region}.json')
    if not skip_hotspots: manager.update_hotspots(hotspots_region)
    if not skip_ladsweb: manager.download_viirs()
    if not skip_preprocess: manager.preprocess_dataset()
    if not skip_getpreds: manager.get_preds(weight_files, threshold=threshold, max_size=max_size)
    if not skip_postprocess: manager.postprocess(filename=f'ba_{manager.time.strftime("%Y%m%d")}',
                    threshold=threshold, area_epsg=area_epsg)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_models.ipynb (unless otherwise specified).

__all__ = ['DROP', 'BTNorm2d', 'LSTM', 'SpaceConv', 'UpSpaceConv', 'TimeConv', 'UpTimeConv', 'SpaceTimeConv',
           'UpSpaceTimeConv', 'BA_Net']

# Cell
from fastai.vision.all import *

# Cell
DROP = 0.2 # Dropout

# Cell
class BTNorm2d(Module):
    "BatchTimeNorm2d module."
    def __init__(self, channels):
        self.bn = nn.BatchNorm2d(channels)

    def forward(self, x):
        n_sequences, ch, sequence_len, sz1, sz2 = x.size()
        x = x.permute(0,2,1,3,4).contiguous().view(n_sequences*sequence_len, ch, sz1, sz2)
        x = self.bn(x)
        x = x.view(n_sequences, sequence_len, ch, sz1, sz2).permute(0,2,1,3,4).contiguous()
        return x

class LSTM(Module):
    "LSTM module."
    def __init__(self, ni, nf):
        self.lstm = nn.LSTM(ni, nf, num_layers=1, bidirectional=False, batch_first=True)

    def forward(self, x):
        bs, ch, ts, sz1, sz2 = x.size()
        x = x.permute(0, 3, 4, 2, 1).contiguous().view(bs*sz1*sz2, ts, ch)
        x, (h, c) = self.lstm(x)
        x = x.view(bs, sz1, sz2, ts, ch).permute(0, 4, 3, 1, 2).contiguous()
        return x, h

class SpaceConv(Module):
    "SpaceConv module."
    def __init__(self, in_ch, out_ch, kernel_sz, stride):
        spaceConv = nn.Conv3d(in_ch, out_ch, kernel_size=(1, kernel_sz, kernel_sz),
                              stride=(1, stride, stride),
                              padding=(0, kernel_sz//2, kernel_sz//2), bias=False)
        layers = [spaceConv, BTNorm2d(out_ch), nn.ReLU(inplace=True)]
        self.conv = nn.Sequential(*layers)
    def forward(self, x): return self.conv(x)

class UpSpaceConv(Module):
    "UpSpaceConv module."
    def __init__(self, in_ch, out_ch):
        upConv = nn.ConvTranspose3d(in_ch, out_ch, kernel_size=(1, 2, 2), stride=(1, 2, 2), bias=False)
        layers = [upConv, BTNorm2d(out_ch), nn.ReLU(inplace=True)]
        self.conv = nn.Sequential(*layers)
    def forward(self, x): return self.conv(x)

class TimeConv(Module):
    "TimeConv module."
    def __init__(self, in_ch, out_ch, kernel_sz, stride, use_lstm):
        self.use_lstm = use_lstm
        self.timeConv = nn.Conv3d(out_ch, out_ch, kernel_size=(kernel_sz, 1, 1),
                            stride=(stride, 1, 1),
                            padding=(kernel_sz//2, 0, 0), bias=False)
        if self.use_lstm: self.timeLstm = LSTM(out_ch, out_ch)
        self.bn = BTNorm2d(out_ch)
        self.relu = nn.ReLU(inplace=True)
    def forward(self, x):
        if self.use_lstm:
            x, _ = self.timeLstm(x)
        x = self.timeConv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x

class UpTimeConv(Module):
    "UpTimeConv module."
    def __init__(self, in_ch, out_ch):
        upTimeConv = nn.ConvTranspose3d(in_ch, out_ch, kernel_size=(2, 1, 1), stride=(2, 1, 1), bias=False)
        layers = [upTimeConv, BTNorm2d(out_ch), nn.ReLU(inplace=True)]
        self.conv = nn.Sequential(*layers)
    def forward(self, x): return self.conv(x)

class SpaceTimeConv(Module):
    "SpaceTimeConv module."
    def __init__(self, in_ch, out_ch, kernel_sz, time_sz, stride, time_stride, time_ch, use_lstm=False):
        spaceConv = SpaceConv(in_ch, out_ch, kernel_sz, stride)
        timeConv = TimeConv(out_ch, out_ch, time_sz, time_stride, use_lstm)
        layers = [spaceConv, timeConv, nn.Dropout3d(DROP)]
        self.stconv = nn.Sequential(*layers)
    def forward(self, x): return self.stconv(x)

class UpSpaceTimeConv(Module):
    "UpSpaceTimeConv module."
    def __init__(self, in_ch, out_ch, time_ch):
        upSpaceConv = UpSpaceConv(in_ch, out_ch)
        upTimeConv = UpTimeConv(out_ch, out_ch)
        layers = [upSpaceConv, upTimeConv, nn.Dropout3d(DROP)]
        self.upstconv = nn.Sequential(*layers)
    def forward(self, x): return self.upstconv(x)

# Cell
class BA_Net(Module):
    "BA-Net model."
    def __init__(self, in_ch, n_classes, sequence_len):
        n=1
        self.stconv1 = SpaceTimeConv(in_ch, n*32, 7, 7, 1, 1, sequence_len, use_lstm=True)
        self.stconv2 = SpaceTimeConv(n*32, n*64, 3, 7, 2, 2, sequence_len//2)
        self.stconv3 = SpaceTimeConv(n*64, n*128, 3, 5, 2, 2, sequence_len//4)
        self.stconv4 = SpaceTimeConv(n*128,n*256, 3, 3, 2, 2, sequence_len//8)
        self.stconv5 = SpaceTimeConv(n*256,n*256, 3, 3, 2, 2, sequence_len//16)
        self.stconv6 = SpaceConv(n*256, n*256, 3, 2)
        self.ustconv6 = UpSpaceConv(n*256, n*256)
        self.ustconv5 = UpSpaceTimeConv(n*512, n*256, sequence_len//8)
        self.ustconv4 = UpSpaceTimeConv(n*512, n*128, sequence_len//4)
        self.ustconv3 = UpSpaceTimeConv(n*256, n*64, sequence_len//2)
        self.ustconv2 = UpSpaceTimeConv(n*128, n*32, sequence_len)
        self.conv3 = nn.Conv3d(n*64, n*64, kernel_size=(3,1,1), padding=(3//2, 0, 0), bias=False)
        self.bn = BTNorm2d(n*64)
        self.relu = nn.ReLU(inplace=True)
        self.final_conv = nn.Conv3d(n*64, n_classes, kernel_size=1)

    def forward(self, x):
        c1 = self.stconv1(x)
        c2 = self.stconv2(c1)
        c3 = self.stconv3(c2)
        c4 = self.stconv4(c3)
        c5 = self.stconv5(c4)
        c6 = self.stconv6(c5)
        up = self.ustconv6(c6)
        up = torch.cat((c5, up), dim=1)
        up = self.ustconv5(up)
        up = torch.cat((c4, up), dim=1)
        up = self.ustconv4(up)
        up = torch.cat((c3, up), dim=1)
        up = self.ustconv3(up)
        up = torch.cat((c2, up), dim=1)
        up = self.ustconv2(up)
        up = torch.cat((c1, up), dim=1)
        up = self.final_conv(self.relu(self.bn(self.conv3(up))))
        return up


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04b_nrt.ipynb (unless otherwise specified).

__all__ = ['RunManager']

# Cell
import pandas as pd
import numpy as np
import scipy.io as sio
import rasterio
import requests
import IPython
import matplotlib.pyplot as plt
from nbdev.imports import test_eq
import datetime
from geoget.download import run_all, Ladsweb
from .core import filter_files, ls, Path, InOutPath, ProjectPath
from .geo import Region
from .data import *
from .predict import predict_nrt
from .web import array2png
from fire_split.core import split_fires, save_data, to_polygon
import pdb
Path.ls = ls

# Cell
class RunManager():
    def __init__(self, project_path:ProjectPath, region: str, time='today',
                 product:str='VIIRS750', days=64):
        self.path    = project_path
        self.time    = self.init_time(time)
        self.product = product
        self.region  = region
        self.days    = days

    @property
    def R(self):
         return Region.load(f'{self.path.config}/R_{self.region}.json')

    def init_time(self, time):
        if time == 'today':
            time = pd.Timestamp(datetime.date.today())
        elif time == 'yesterday':
            time = pd.Timestamp(datetime.date.today())-pd.Timedelta(days=1)
        return time

    def last_n_days(self, time:pd.Timestamp, days):
        return pd.date_range(start=time-pd.Timedelta(days=days-1), periods=days,
                              freq='D')

    def check_data(self):
        "Check existing and missing files in dataset folder."
        times = self.last_n_days(self.time, self.days)
        files, missing_files = [], []
        for t in times:
            tstr = t.strftime('%Y%m%d')
            file = self.path.dataset/f'{self.product}{self.region}_{tstr}.nc'
            if file.is_file():
                files.append(file)
            else:
                missing_files.append(file)
        return {'files': files, 'missing_files': missing_files}

    def get_download_dates(self):
        "Find for which new dates the files need to be downloaded."
        all_times = self.last_n_days(self.time, self.days)
        files = self.check_data()['files']
        existing_times = [pd.Timestamp(Path(o).stem.split('_')[-1]) for o in files]
        download_times = set(all_times) - set(existing_times)
        if len(download_times) > 0:
            start = min(download_times).strftime('%Y-%m-%d 00:00:00')
            end = max(download_times).strftime('%Y-%m-%d 23:59:59')
        else:
            start, end = None, None
        return start, end

    def update_hotspots(self, location, mode='7d', save=True):
        """Update hotspots file with new data.
            location is according to the data url naming format
            mode can be on of: 24h, 48h, 7d"""
        url1 = f'https://firms.modaps.eosdis.nasa.gov/' \
            f'data/active_fire/suomi-npp-viirs-c2/csv/SUOMI_VIIRS_C2_{location}_{mode}.csv'
        url2 = f'https://firms.modaps.eosdis.nasa.gov/' \
            f'data/active_fire/noaa-20-viirs-c2/csv/J1_VIIRS_C2_{location}_{mode}.csv'
        files = self.path.hotspots.ls(include=['.csv', f'hotspots{self.region}'])
        frp = [pd.read_csv(f) for f in files]
        frp = pd.concat([*frp, *[pd.read_csv(url) for url in [url1, url2]]], axis=0,
                        sort=False).drop_duplicates().reset_index(drop=True)
        if save:
            frp.to_csv(self.path.hotspots/f'hotspots{self.region}.csv', index=False)
            print(f'hotspots{self.region}.csv updated')
        else: return frp

    def download_viirs(self, products=['VJ102IMG', 'VJ103IMG'], collection='5201', daynight='D'):
        "Download viirs data needed for the dataset."
        tstart, tend = self.get_download_dates()
        if tstart is not None:
            for product in products:
                print(product)
                lads = Ladsweb(
                    product=product,
                    collection=collection,
                    tstart=tstart,
                    tend=tend,
                    bbox=list(self.R.bbox), # left bottom right top
                    daynight=daynight) # D N DNB
                lads.download_raw_files(self.path.ladsweb)

    def preprocess_dataset(self, max_workers=1, replace=False):
        if self.product == 'VIIRS375':
            bands = ['I01', 'I02', 'I03', 'I04', 'I05', 'solar_zenith', 'sensor_zenith', 'latitude', 'longitude']
            viirs = Viirs375Dataset(InOutPath(self.path.ladsweb, self.path.dataset), self.R, bands=bands)
            rename1 = BandsRename(
                ['I01', 'I02', 'I03', 'I04', 'I05', 'solar_zenith', 'sensor_zenith', 'latitude', 'longitude'],
                ['Reflectance_I1', 'Reflectance_I2', 'Reflectance_I3', 'Radiance_I4', 'Radiance_I5',
                 'SolarZenithAngle', 'SatelliteZenithAngle', 'Latitude', 'Longitude'])
            merge_tiles = MergeTiles('SatelliteZenithAngle')
            mir_calc = MirCalc('SolarZenithAngle', 'Radiance_I4', 'Radiance_I5')
            rename2 = BandsRename(['Reflectance_I1', 'Reflectance_I2'], ['Red', 'NIR'])
            bfilter = BandsFilter(['Red', 'NIR', 'MIR'])
            act_fires = ActiveFiresLog(f'{self.path.hotspots}/hotspots{self.region}.csv')
            viirs.process_all(proc_funcs=[mask2nan, rename1, merge_tiles, mir_calc,
                                          rename2, bfilter, act_fires, nan2zero],
                              max_workers=max_workers, replace=replace)
        else: raise NotImplementedError(f'Not implemented for {self.product}.')

    def init_model_weights(self, weight_files:list):
        "Downloads model weights if they don't exist yet on config directory."
        local_files = []
        for w in weight_files:
            file_save = self.path.config/w
            if not file_save.is_file():
                print(f'Downloading model weights {w}')
                url = f'https://github.com/mnpinto/banet_weights/raw/master/model/{w}'
                file = requests.get(url)
                open(str(file_save), 'wb').write(file.content)
            local_files.append(file_save)
        return local_files

    def get_preds(self, weight_files:list, threshold=0.5, save=True, max_size=2000):
        "Computes BA-Net predictions ensembling the models in the weight_files list."
        local_files = self.init_model_weights(weight_files)
        iop = InOutPath(self.path.dataset, self.path.outputs, mkdir=False)
        region = self.R.new()
        predict_nrt(iop, self.time, local_files, region, threshold=threshold,
                    save=save, max_size=max_size, product=self.product)

    def postprocess(self, filename, threshold=0.5, interval_days=2, interval_pixels=2,
                    min_size_pixels=25, area_epsg=None, keys=['burned', 'date'],
                    geotiff_only=False, padding=1):
        "Computes tifs and shapefiles from outputs."
        data = sio.loadmat(self.path.outputs/f'{filename}.mat')
        times = pd.DatetimeIndex([pd.Timestamp(str(o)) for o in data['times']])
        unique_years = times.year.unique()
        if len(unique_years) > 2: raise NotImplementedError(
            'Not implemented for more than 2 years range')
        burned = data[keys[0]]
        date = data[keys[1]]
        I = burned < threshold
        date[I] = np.nan
        It = date[date>=0].astype(np.uint16)
        if len(unique_years) > 1:
            m = (times.year[It] == unique_years[1])*pd.Timestamp(f'{unique_years[1]}-12-31').dayofyear
        else: m = 0
        date[date>=0] = times[It].dayofyear + m
        date[np.isnan(date)] = 0
        date = date.astype(np.uint16)
        burned = (burned*255).astype(np.uint8)
        region = self.R.new()
        if geotiff_only:
            raster = np.array([burned, date])
            save_data(self.path.web/f'{filename}.tif', raster, crs=region.crs,
                                   transform=region.transform)
        else:
            labels, df = split_fires(date, interval_days=interval_days,
                         interval_pixels=interval_pixels,
                         min_size_pixels=min_size_pixels)
            burned[labels==0] = 0
            date[labels==0] = 0
            raster = np.array([burned, date])
            save_data(self.path.web/f'{filename}.tif', raster, crs=region.crs,
                                   transform=region.transform)
            for i in range(labels.max()):
                l = (i+1)
                f = labels == l
                args = np.argwhere(f)
                lon, lat = region.coords()
                (rmin, cmin), (rmax, cmax) = args.min(0), args.max(0)
                rmax += 1
                cmax += 1
                rmin = max(rmin, 1)
                cmin = max(cmin, 1)
                lat_r = lat[rmin-padding:rmax+padding]
                lon_r = lon[cmin-padding:cmax+padding]
                tfm = rasterio.Affine(region.pixel_size, 0, lon_r.min(), 0, -region.pixel_size, lat_r.max())
                burned_r = data[keys[0]][rmin-padding:rmax+padding, cmin-padding:cmax+padding].copy().astype(np.float16)
                date_r =  data[keys[1]][rmin-padding:rmax+padding, cmin-padding:cmax+padding].copy().astype(np.float16)
                burned_r[f[rmin-padding:rmax+padding, cmin-padding:cmax+padding]==0] = np.nan
                date_r[f[rmin-padding:rmax+padding, cmin-padding:cmax+padding]==0] = np.nan

                burned_r = burned_r*255
                burned_r[np.isnan(burned_r)] = 0
                burned_r = burned_r.astype(np.uint16)
                date_r[np.isnan(date_r)] = 0
                date_r = date_r.astype(np.uint16)
                raster = np.array((burned_r, date_r))

                (self.path.web/'events').mkdir(exist_ok=True)
                save_data(self.path.web/'events'/f'{filename}_{l}.tif',
                          raster, crs=region.crs, transform=tfm)
                im = array2png(burned_r, cmap='RdYlGn_r')
                im.save(self.path.web/'events'/f'{filename}_{l}_cl.png')
                im = array2png(date_r, cmap='jet')
                im.save(self.path.web/'events'/f'{filename}_{l}_bd.png')

            df = to_polygon(labels, region.crs, region.transform, df, area_epsg=area_epsg)
            df['area_ha'] = df['area_ha'].astype(np.uint32)
            df.to_file(self.path.web/f'{filename}.shp')
            df.to_file(self.path.web/f'{filename}.json', driver='GeoJSON')


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07_web.ipynb (unless otherwise specified).

__all__ = ['array2png']

# Cell
import PIL
import matplotlib
import numpy as np

# Cell
def array2png(data, cmap='RdYlGn_r'):
    "Convert uint8 2d-array 'data' to png using colormap 'cmap'"
    mask = data == 0
    unique = sorted(np.unique(data[data>0]).reshape(-1).tolist())
    cmap = matplotlib.cm.get_cmap(cmap)
    colors = (cmap(np.linspace(0,1,unique[-1]-unique[0]+1))*255).astype(np.uint8)
    data[data>0] -= unique[0]
    data = data.astype(np.uint8)
    data = colors[data]
    data[mask, -1] = 0
    return PIL.Image.fromarray(data)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/04c_historical.ipynb (unless otherwise specified).

__all__ = ['RunManager']

# Cell
import pandas as pd
import numpy as np
import scipy.io as sio
import rasterio
import requests
import IPython
import matplotlib.pyplot as plt
from nbdev.imports import test_eq
import datetime
from geoget.download import run_all
import banet.nrt
from .core import filter_files, ls, Path, InOutPath, ProjectPath
from .geo import Region
from .data import *
from .predict import predict_time
from fire_split.core import split_fires, save_data, to_polygon
Path.ls = ls

# Cell
class RunManager(banet.nrt.RunManager):
    def __init__(self, project_path:ProjectPath, region:str, times:pd.DatetimeIndex,
                 product:str='VIIRS750'):
        """
        project_path: banet.core.ProjectPath object
        region: name of the region
        times: dates for the first day of month for each month to use
        product: VIIRS750 or VIIRS375
        """
        self.path    = project_path
        self.times   = self.init_times(times)
        self.product = product
        self.region  = region

    def init_times(self, times):
        tstart = times[0] - pd.Timedelta(days=15)
        tstart = pd.Timestamp(f'{tstart.year}-{tstart.month}-01')
        tend = times[-1] + pd.Timedelta(days=75)
        tend = pd.Timestamp(f'{tend.year}-{tend.month}-01') - pd.Timedelta(days=1)
        return pd.date_range(tstart, tend, freq='D')

    def check_data(self):
        "Check existing and missing files in dataset folder."
        times = self.times
        files, missing_files = [], []
        for t in times:
            tstr = t.strftime('%Y%m%d')
            file = self.path.dataset/f'{self.product}{self.region}_{tstr}.nc'
            if file.is_file():
                files.append(file)
            else:
                missing_files.append(file)
        return {'files': files, 'missing_files': missing_files}

    def get_download_dates(self):
        "Find for which new dates the files need to be downloaded."
        files = self.check_data()['files']
        if len(files) == 0:
            start = self.times[0]
        else:
            start = pd.Timestamp(files[-1].stem.split('_')[-1])+pd.Timedelta(days=1)
        start = start.strftime('%Y-%m-%d 00:00:00')
        end = self.times[-1].strftime('%Y-%m-%d 23:59:59')
        return start, end

    def download_viirs(self, maxOrderSize=[1800, 1200]):
        "Download viirs data needed for the dataset."
        tstart, tend = self.get_download_dates()
        region = self.R.new()

        if self.product == 'VIIRS750':
            viirs_downloader = VIIRS750_download(region, tstart, tend)
            viirs_downloader_list = viirs_downloader.split_times()

        elif self.product == 'VIIRS375':
            viirs_downloader1 = VIIRS375_download(region, tstart, tend)
            region.pixel_size = 0.1 # Angles can be interpolated later
            viirs_downloader2 = VIIRS750_download(region, tstart, tend,
                                bands=['SolarZenithAngle', 'SatelliteZenithAngle'])
            viirs_downloader_list1 = viirs_downloader1.split_times(maxOrderSize=maxOrderSize[0])
            viirs_downloader_list2 = viirs_downloader2.split_times(maxOrderSize=maxOrderSize[1])
            viirs_downloader_list = [*viirs_downloader_list1, *viirs_downloader_list2]
        else: raise NotImplementedError(f'Not implemented for {self.product}.')

        run_all(viirs_downloader_list, self.path.ladsweb)

    def get_preds(self, weight_files:list, threshold=0.5, save=True, max_size=2000,
                  filename='data', check_file=False, verbose=False):
        "Computes BA-Net predictions ensembling the models in the weight_files list."
        local_files = self.init_model_weights(weight_files)
        iop = InOutPath(self.path.dataset, self.path.outputs, mkdir=False)
        region = self.R.new()
        predict_time(iop, self.times, local_files, region, threshold=threshold,
                     save=save, max_size=max_size, product=self.product, output=filename,
                      check_file=check_file, verbose=verbose)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_geo.ipynb (unless otherwise specified).

__all__ = ['Region', 'open_shp', 'open_tif', 'bounds_from_shapefile', 'size_from_bounds', 'size_resolution_assert',
           'rasterize', 'downsample', 'is_intersection', 'polygon_from_bounds', 'crop', 'bounds_from_coords']

# Cell
import geopandas as gp
import numpy as np
import matplotlib.pyplot as plt
import json
import numpy as np
import pandas as pd
import shapely
import rasterio
import rasterio.features
from rasterio import features
from rasterio.mask import mask
from rasterio.merge import merge
from rasterio.coords import BoundingBox
from rasterio.crs import CRS
import warnings
from fastcore.test import *

from .core import *

# Cell
class Region():
    """Defines a geographical region with a name, a bounding box and the pixel size"""
    def __init__(self, name:str, bbox:list, pixel_size:float, epsg:int=4326,
                 shape=None):
        self.name       = name
        self.bbox       = rasterio.coords.BoundingBox(*bbox) # left, bottom, right, top
        self.pixel_size = pixel_size
        self.epsg       = epsg
        self._shape     = shape

    @property
    def width(self):
        "Width of the region"
        if self._shape is None:
            return np.arange(self.bbox.left, self.bbox.right, self.pixel_size).shape[0]
        else: return self.shape[1]

    @property
    def height(self):
        "Height of the region"
        if self._shape is None:
            return np.arange(self.bbox.bottom, self.bbox.top, self.pixel_size).shape[0]
        else: return self.shape[0]

    @property
    def transform(self):
        "Rasterio Affine transform of the region"
        return rasterio.transform.from_bounds(*self.bbox, self.width, self.height)

    @property
    def crs(self):
        return CRS.from_epsg(self.epsg)

    @property
    def shape(self):
        "Shape of the region (height, width)"
        if self._shape is None:
            return (self.height, self.width)
        else: return self._shape

    def coords(self, offset='ul'):
        "Computes longitude and latitude arrays given a shape and a rasterio Affine transform"
        rxy = rasterio.transform.xy
        ys, xs = map(range, self.shape)
        return (np.array(rxy(self.transform, [0]*len(xs), xs, offset=offset)[0]),
                np.array(rxy(self.transform, ys, [0]*len(ys), offset=offset)[1]))

    @classmethod
    def load(cls, file):
        "Loads region information from json file"
        with open(file, 'r') as f:
            args = json.load(f)
        return cls(args['name'], args['bbox'], args['pixel_size'])

    def new(self, name:str=None, bbox:list=None, pixel_size:float=None, epsg:int=None,
            shape:tuple=None):
        "Create new region with updated parameters."
        if name is None: name = self.name
        if bbox is None: bbox = list(self.bbox)
        if pixel_size is None: pixel_size = self.pixel_size
        if epsg is None: epsg = self.epsg
        return Region(name, bbox, pixel_size, epsg=epsg, shape=shape)

    def export(self, file):
        """Exports region information to json file"""
        dict2json(self.__dict__, file)

    def __repr__(self):
        return '\n'.join([f'{i}: {o}' for i, o in self.__dict__.items()]) + '\n'

# Cell
def open_shp(file):
    "Read shapefile"
    return gp.read_file(file)

def open_tif(file):
    "Read tiff"
    return rasterio.open(file)

def bounds_from_shapefile(shapefile):
    "Computes bounding box for shapefile"
    bounds = shapefile.bounds
    return bounds.minx.min(), bounds.miny.min(), bounds.maxx.max(), bounds.maxy.max()

def size_from_bounds(bounds, resolution):
    "Computes width and height from bounds for a given pixel resolution"
    mlon = np.mean([bounds[2], bounds[0]])
    width = np.ceil((bounds[2]-bounds[0])*(111100/resolution)*np.cos(np.deg2rad(mlon))).astype(int)
    height = np.ceil((bounds[3]-bounds[1])*(111100/resolution)).astype(int)
    return width, height

def size_resolution_assert(size, resolution):
    if size is None and resolution is None:
        raise Exception('You must define either size or resolution')
    if size is not None and resolution is not None:
        warnings.warn('resolution not used, computed based on size and bounds')

def rasterize(x, value_key=None, region=None, merge_alg='replace'):
    "Rasterize shapefile"
    if merge_alg == 'replace':
        merge_alg = rasterio.enums.MergeAlg.replace
    elif merge_alg == 'add':
        merge_alg = rasterio.enums.MergeAlg.add
    values = [1]*len(x) if value_key is None else x[value_key]
    shapes = (v for v in zip(x.geometry, values))
    return rasterio.features.rasterize(shapes, out_shape=region.shape,
            transform=region.transform, merge_alg=merge_alg)

def downsample(x, src_tfm=None, dst_tfm=None, dst_shape=None,
               src_crs={'init': 'EPSG:4326'}, dst_crs={'init': 'EPSG:4326'},
               resampling='average'):
    "Donwsample a numpy array x"
    if resampling == 'average':
        resampling = rasterio.warp.Resampling.average
    elif resampling == 'bilinear':
        resampling = rasterio.warp.Resampling.bilinear
    elif resampling == 'nearest':
        resampling = rasterio.warp.Resampling.nearest
    out = np.zeros(dst_shape)
    rasterio.warp.reproject(x, out, src_transform=src_tfm, dst_transform=dst_tfm,
                            src_crs=src_crs, dst_crs=dst_crs, resampling=resampling)
    return out

def is_intersection(gdf1, gdf2):
    "Find the intersection between two geo pandas dataframes"
    return len(gp.overlay(gdf1, gdf2, how='intersection')) > 0

def polygon_from_bounds(bounds, to_GeoDataFrame=False, crs={'init': 'EPSG:4326'}):
    "Create a polygon object from bounds"
    b_ind = [[0,1],[2,1],[2,3],[0,3]]
    shape = shapely.geometry.Polygon([(bounds[i],bounds[j]) for i, j in b_ind])
    if to_GeoDataFrame: shape = gp.GeoDataFrame(crs=crs, geometry=[shape])
    return shape

def crop(x, bounds=None, shape=None, crop=True):
    """
    Crop rasterio dataset for a region defined by bounds.
        x is a dataset or a list of datasets (rasterio.open).
        If list then merge with bounds is used.
        else mask is used to crop given bounds or any given shape.
    """
    if len(x) == 1 and isinstance(x, list):
        x = x[0]
    if isinstance(x, list):
        out, transform = merge(x, bounds)
    else:
        if bounds is not None: shape = polygon_from_bounds(bounds)
        out, transform = mask(x, shapes=[shape], crop=crop)
    return out.squeeze(), transform

def bounds_from_coords(lon, lat):
    "Compute bounds list form lon lat coords"
    return lon.min(), lat.min(), lon.max(), lat.max()


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02_data.ipynb (unless otherwise specified).

__all__ = ['VIIRS375_download', 'VIIRS750_download', 'BandsFilter', 'BandsRename', 'MergeTiles', 'BandsAssertShape',
           'ActiveFiresLog', 'InterpolateAngles', 'MirCalc', 'BaseDataset', 'mask2nan', 'nan2zero', 'group_files',
           'ViirsDataset', 'Viirs375Dataset', 'Viirs750Dataset', 'MCD64Dataset', 'FireCCI51Dataset', 'AusCoverDataset',
           'MTBSDataset', 'ICNFDataset', 'Region2Tiles']

# Cell
import numpy as np
import pandas as pd
import re
import gc
import sys
from pathlib import Path
from rasterio.coords import disjoint_bounds
from tqdm import tqdm
import scipy.io as sio
from functools import partial
import netCDF4
from netCDF4 import Dataset
from pyhdf.SD import SD, SDC
from geopandas import GeoDataFrame
from shapely.geometry import Point
import matplotlib.pyplot as plt
from concurrent.futures import ThreadPoolExecutor
from warnings import warn
from nbdev.imports import test_eq
from pyresample.geometry import SwathDefinition, AreaDefinition, CRS, create_area_def
from pyresample import kd_tree

from geoget.download import *
from .core import *
from .geo import *
from IPython.core.debugger import set_trace

# Cell
class VIIRS375_download(Ladsweb):
    "Utility for downloading VIIRS 375m data to create the dataset."
    def __init__(self, region, tstart, tend, bands=None):
        product = 'NPP_VIAES_L1'
        collection = '5000'
        if bands is None:
            bands = ['Reflectance_I1', 'Reflectance_I2', 'Reflectance_I3',
                      'Radiance_I4', 'Radiance_I5']
        super().__init__(product, collection, tstart, tend, list(region.bbox),
                         bands, daynight='D', repPixSize=region.pixel_size)

# Cell
class VIIRS750_download(Ladsweb):
    "Utility for downloading VIIRS 750m data to create the dataset."
    def __init__(self, region, tstart, tend, bands=None):
        product = 'NPP_VMAES_L1'
        collection = '5000'
        if bands is None:
            bands = ['Reflectance_M5', 'Reflectance_M7', 'Reflectance_M10',
                      'Radiance_M12', 'Radiance_M15', 'SolarZenithAngle',
                      'SatelliteZenithAngle']
        super().__init__(product, collection, tstart, tend, list(region.bbox),
                         bands, daynight='D', repPixSize=region.pixel_size)

# Cell
class BandsFilter():
    """Remove bands not in to_keep list from the dictionary."""
    def __init__(self, to_keep: list):
        self.to_keep = to_keep if isinstance(to_keep, list) else [to_keep]

    def __call__(self, data:dict, *args, **kwargs) -> dict:
        keys = [k for k in data]
        for k in keys:
            if k not in self.to_keep:
                del data[k]
        return data

# Cell
class BandsRename():
    def __init__(self, input_names:list, output_names:list):
        self.input_names = input_names if isinstance(input_names, list) else [input_names]
        self.output_names = output_names if isinstance(output_names, list) else [output_names]

    def __call__(self, data:dict, *args, **kwargs) -> dict:
        for i, o in zip(self.input_names, self.output_names):
            data[o] = data.pop(i)
        return data

# Cell
class MergeTiles():
    def __init__(self, band:str, ignore:list=[]):
        self.band = band
        self.ignore = ignore

    def __call__(self, data:dict, *args, **kwargs) -> dict:
        d = np.nanmean(np.array(data[self.band]), axis=(1,2))
        d = np.array(np.array(d).argsort())
        masks = np.array(data[self.band])[d]
        for k in data:
            if k not in self.ignore:
                data_aux = np.zeros_like(data[k][0])*np.nan
                for dband, mask in zip(np.array(data[k])[d], masks):
                    I = (np.isnan(data_aux)) & (~np.isnan(mask))
                    data_aux[I] = dband[I]
                data[k] = data_aux
        return data

# Cell
class BandsAssertShape():
    def __call__(self, data:dict, max_size=None, *args, **kwargs) -> dict:
        for k in kwargs['cls'].bands:
            rshape = kwargs['cls'].region.shape
            if isinstance(data[k], list):
                for d in data[k]:
                    shape = d.shape
                    if len(shape) == 3: # first is time
                        shape = shape[1:]
                    if shape != rshape:
                        error = f'{k} shape {shape} does not match region shape {rshape}'
                        raise Exception(error)
            else:
                shape = data[k].shape
                if len(shape) == 3: # first is time
                    shape = shape[1:]
                if shape != rshape:
                    error = f'{k} shape {shape} does not match region shape {rshape}'
                    raise Exception(error)
        return data

# Cell
class ActiveFiresLog():
    """Get active fires, interpolate to grid and apply log1p."""
    def __init__(self, file):
        self.file = file
        self.lon = None
        self.lat = None
        self.df = self.load_csv()

    def load_csv(self):
        return pd.read_csv(self.file, parse_dates=['acq_date']).set_index('acq_date')

    def __call__(self, data, time, *args, **kwargs):
        R = kwargs['cls'].region.new()
        if "R" in data and isinstance(data["R"], Region):
            R = data["R"]
        if R.pixel_size < 0.01:
            Rbase = R.new(pixel_size=0.01)
            self.lon, self.lat = Rbase.coords()
        else:
            self.lon, self.lat = R.coords()
        frp = self.df[self.df.index == time]

        if len(frp) > 0:
            geometry = [Point(xy) for xy in zip(frp['longitude'], frp['latitude'])]
            frp = GeoDataFrame(frp, geometry=geometry)
            out = rasterize(frp, 'frp', Rbase, merge_alg='add')
            if R.pixel_size < 0.01:
                out = downsample(out, src_tfm=Rbase.transform, dst_tfm=R.transform,
                                 dst_shape=R.shape, resampling='bilinear')
            out[out==0] = np.nan
        else: out = np.zeros(R.shape)*np.nan
        data['FRP'] = np.log1p(out)
        return data

# Cell
class InterpolateAngles():
    """Interpolated Angles to working resolution."""
    def __init__(self, R_base:Region, R_interp:Region, bands:list):
        self.R_base   = R_base
        self.R_interp = R_interp
        self.bands    = bands

    def __call__(self, data:dict, *args, **kwargs):
        R_interp = self.R_interp
        if "R" in data and isinstance(data["R"], Region):
            R_interp = data["R"]
        for b in self.bands:
            data[b] = [downsample(data[b][i], src_tfm=self.R_base.transform,
                                 dst_tfm=R_interp.transform,
                                 dst_shape=R_interp.shape,
                                 resampling='bilinear')
                       for i in range(len(data[b]))]
        return data

# Cell
class MirCalc():
    def __init__(self, solar_zenith_angle:str, mir_radiance:str, tir_radiance:str,
                 output_name:str='MIR'):
        self.sza = solar_zenith_angle
        self.r_mir = mir_radiance
        self.r_tir = tir_radiance
        self.output_name = output_name

    def __call__(self, data:dict, *args, **kwargs):
        sza = data[self.sza]
        mir = data[self.r_mir]
        tir = data[self.r_tir]
        data[self.output_name] = self.refl_mir_calc(mir, tir, sza, sensor=kwargs['cls'].name)
        return data

    def refl_mir_calc(self, mir, tir, sza, sensor):
        """
        Computes the MIR reflectance from MIR radiance and Longwave IR radiance.
        sensor can be "VIIRS375" or "VIIRS750"
        sza is the solar zenith angle
        for VIIRS375, mir is band I4 and tir band I5
        for VIIRS750, mir is band M12 and tir band M15
        returns a matrix of MIR reflectances with the same shape as mir and tir inputs.
        Missing values are represented by 0.
        """
        lambda_M12= 3.6966
        lambda_M15=10.7343
        lambda_I4 = 3.7486
        lambda_I5 = 11.4979

        c1 = 1.1911e8 # [ W m-2 sr-1 (micrometer -1)-4 ]
        c2 = 1.439e4 # [ K micrometer ]
        E_0_mir_M12 = 11.7881 # M12 newkur_semcab
        E_0_mir_I4= 11.2640 # I4 newkur_semcab

        if sensor=='VIIRS375':
            lambda_mir = lambda_I4
            lambda_tir = lambda_I5
            E_0_mir = E_0_mir_I4
        elif sensor=='VIIRS750':
            lambda_mir = lambda_M12
            lambda_tir = lambda_M15
            E_0_mir = E_0_mir_M12
        else: raise NotImplementedError(
            f'refl_mir_calc not implemented for {sensor}. Available options are VIIRS750 and VIIRS375.')

        miu_0=np.cos((sza*np.pi)/180)

        mir[mir <= 0] = np.nan
        tir[tir <= 0] = np.nan

        # Brighness temperature
        a1 = (lambda_tir**5)
        a = c1/(a1*tir)
        logaritmo = np.log(a+1)
        divisor = lambda_tir*logaritmo
        T = (c2/divisor)
        del a, logaritmo, divisor

        # Plank function
        divisor2 = (lambda_mir*T)
        exponencial = np.exp(c2/divisor2)
        b = c1*(lambda_mir**-5)
        BT_mir = b/(exponencial-1)
        del divisor2, exponencial, b, T

        # MIR reflectance
        c = (E_0_mir*miu_0)/np.pi
        termo1 = (mir-BT_mir)
        termo2 = (c-BT_mir)
        Refl_mir = termo1/termo2
        Refl_mir[Refl_mir <= 0] = 0
        return Refl_mir

# Cell
class BaseDataset():
    def __init__(self, name:str, paths:InOutPath, region:Region,
                 times:pd.DatetimeIndex=None, bands:list=None):
        self.paths = paths
        self.region = region
        self.name = name
        self.times = times
        self.bands = bands

        if self.times is None:
            self.times = self.find_dates()

    def list_files(self, time:pd.Timestamp) -> list:
        "This method should return a list of filenames corresponding to the given Timestamp."
        pass

    def find_dates(self):
        """This method should return a pd.DatetimeIndex
        with list of dates present in the data available in the input path."""
        pass

    def match_times(self, other, on='month'):
        "Set the times attribute to match the times of other dataset."
        if on != 'month':
            raise NotImplementedError('match_times is only implemented on month.')

        ym_other = sorted(set([(t.year, t.month) for t in other.times]))
        out = []
        for t in self.times:
            if (t.year, t.month) in ym_other:
                out.append(t)
        self.times = pd.DatetimeIndex(out)

    def filter_times(self, year):
        """To select only a specific year. This can be usefull for testing and
           for adding more new years and avoid reprocessing all the dataset."""
        if year is not None:
            self.times = self.times[self.times.year == year]

    def open(self, files:list, crop=None) -> dict:
        """This method is used to open a file or list of files for a given
        time period and returns a dictionary with the data ready to be passed
        to the processing functions."""
        pass

    def save(self, time:pd.Timestamp, data:dict, crop=None):
        "Saves data in a single file for a specific timestamp in netcdf4 format."
        scale_factor = 100
        mask_value = 65535
        tstr = time.strftime('%Y%m%d')
        filename = f'{self.paths.dst}/{self.name}{self.region.name}_{tstr}.nc'
        mode = "r+" if Path(filename).is_file() and crop is not None else "w"
        with netCDF4.Dataset(filename, mode, format="NETCDF4") as rootgrp:
            if mode == "w":
                latD = rootgrp.createDimension("lat", self.region.height)
                lonD = rootgrp.createDimension("lon", self.region.width)
            for k in data:
                if mode == "w":
                    a = rootgrp.createVariable(k,"u2",("lat","lon",), zlib=True)
                    a.setncattr('scale', scale_factor)
                    a.setncattr('mask', mask_value)
                else: a = rootgrp[k]
                v = data[k]*scale_factor
                v[np.isnan(v)] = mask_value
                assert np.sum(data[k] < 0) == 0
                if np.sum(data[k]*scale_factor > mask_value) > 0:
                    warn(f'Clipping data to {mask_value-1} for {k}.')
                    v[v>mask_value] = mask_value-1
                if crop is None:
                    a[:] = v.astype(np.uint16)
                else:
                    a[crop[0]:crop[1],crop[2]:crop[3]] = v.astype(np.uint16)

    def process_one(self, time:pd.Timestamp, proc_funcs:list=[], save=True, max_size=None,
                    **proc_funcs_kwargs):
        """This method defines a processing pipeline consisting of opening the file
        using the `open` method, applying each of the `proc_funcs` to the output of the previous
        and `save` the processed data using save method."""
        tstr = time.strftime('%Y%m%d')
        files = self.list_files(time)
        #try:
        tstr = time.strftime('%Y%m%d')
        filename = f'{self.paths.dst}/{self.name}{self.region.name}_{tstr}.nc'
        if not Path(filename).is_file():
            if len(files) > 0:
                if max_size is None:
                    si = [[0, self.region.shape[0], 0, self.region.shape[1]]]
                else:
                    si = [[max(0,j*max_size),(j+1)*max_size,max(0,i*max_size),(i+1)*max_size]
                           for i in range((self.region.shape[1]-1)//max_size+1)
                           for j in range((self.region.shape[0]-1)//max_size+1)]
                for c in si:
                    #print(f'Processing crop {c}')
                    data = self.open(files, crop=c)
                    kwargs = {'cls': self, **proc_funcs_kwargs}
                    for f in proc_funcs:
                        data = f(data, time, **kwargs)
                    if save:
                        self.save(time, data, crop=c)
                        del data
                        gc.collect()
                    else: return data
            else:
                warn(f'No files for {time}. Skipping to the next time.')
        else: warn(f'Skipping files for {time}. File already exists.')
        #except:
        #    msg = f'Unable to process files for {time}. Check if files are corrupted. Skipping to the next time. { sys.exc_info()[0]}'
        #    warn(msg, UserWarning)

    def process_all(self, proc_funcs=[], max_workers=1, max_size=None, **proc_funcs_kwargs):
        """`process_all` runs `process_one` in parallel using the number of workers defined
        by `max_workers` and passes the `proc_funcs` list to `process_one` method"""
        process_one = partial(self.process_one, proc_funcs=proc_funcs, max_size=max_size,
                              **proc_funcs_kwargs)
        with ThreadPoolExecutor(max_workers) as e:
            list(tqdm(e.map(process_one, self.times), total=len(self.times)))

    def __repr__(self):
        return '\n'.join([f'{i}: {o}' for i, o in self.__dict__.items()]) + '\n'

# Cell
def mask2nan(data:dict, *args, **kwargs) -> dict:
    for k in data:
        k_list = data[k]
        for i, d in enumerate(k_list):
            mask = d.mask
            data[k][i] = d.data
            data[k][i][mask] = np.nan
    return data

def nan2zero(data:dict, *args, **kwargs) -> dict:
    for k in data:
        data[k][np.isnan(data[k])] = 0
    return data

def group_files(files:list):
    return pd.DataFrame({'files': files, 'ids':['.'.join(f.stem.split('.')[-4:-2]) for f in files]}
                        ).groupby('ids').agg(lambda x : list(x)).files.values.tolist()

class ViirsDataset(BaseDataset):
    "Subclass of `BaseDataset` to process VIIRS 750-meter bands."
    _name = None
    def __init__(self, paths:InOutPath, region:Region,
                 times:pd.DatetimeIndex=None, bands:list=None):
        super().__init__(self._name, paths, region, times, bands)
        self.times = self.check_files()

    def list_files(self, time:pd.Timestamp)-> list:
        if time in self.times:
            dayOfYear = str(time.dayofyear).zfill(3)
            files = self.paths.src.ls(include=[f'.A{time.year}{dayOfYear}.', '.nc'])
        return files

    def check_files(self):
        not_missing = []
        for i, t in tqdm(enumerate(self.times), total=len(self.times)):
            files = self.list_files(t)
            if len(files)==0:
                print(f'Missing files for {t}')
            else: not_missing.append(i)
        return self.times[not_missing]

    def find_dates(self, first:pd.Timestamp=None, last:pd.Timestamp=None):
        pattern = r'^\w+.A(20[0-9][0-9])([0-3][0-9][0-9])..*$'
        times = []
        for f in self.paths.src.ls():
            x = re.search(pattern, f.stem)
            if x is not None:
                year, doy = map(x.group, [1,2])
                times.append(pd.Timestamp(f'{year}-01-01') + pd.Timedelta(days=int(doy)-1))
        self.times = pd.DatetimeIndex(sorted(set(times)))
        if first is not None:
            self.times = self.times[self.times>=first]
        if last is not None:
            self.times = self.times[self.times<=last]
        return self.times

    def open_netcdf4(self, files:list, dtype=np.float32) -> dict:
        data_dict = {}
        for f0 in files:
            nc_data = netCDF4.Dataset(str(f0), mode='r')
            if 'observation_data' in nc_data.groups:
                nc_data = nc_data.groups['observation_data']
            elif 'geolocation_data' in nc_data.groups:
                nc_data = nc_data.groups['geolocation_data']
            bands = [k for k in nc_data.variables]
            for s in list(set(bands).intersection(set(self.bands))):
                data = nc_data[s][:].astype(dtype)
                if s in data_dict: raise Exception('...')
                data_dict[s] = data
        mask = data_dict['I01'].mask | data_dict['I02'].mask | data_dict['I03'].mask | data_dict['I04'].mask | data_dict['I05'].mask
        for k in data_dict:
            data_dict[k].mask = mask
        return data_dict

    def resample(self, data: dict, epsg=4326, max_distance_meter=1000, num_workers=8):
        swath_def = SwathDefinition(lons=data['longitude'], lats=data['latitude'])
        crs = CRS(f'EPSG:{epsg}')
        area_def = create_area_def(crs.name, crs.to_dict(), area_extent=self.region.bbox, resolution=self.region.pixel_size)
        valid_input_index, valid_output_index, index_array, distance_array = \
            kd_tree.get_neighbour_info(swath_def, area_def, max_distance_meter, neighbours=1, nprocs=num_workers)
        for k in self.bands:
            data[k] = kd_tree.get_sample_from_neighbour_info('nn', output_shape=self.region.shape, data=data[k],
                                                   valid_input_index=valid_input_index, valid_output_index=valid_output_index,
                                                   index_array=index_array, distance_array=distance_array, fill_value=None)
        return data

    def process_one(self, time:pd.Timestamp, proc_funcs:list=[], save=True, replace=False,
                    **proc_funcs_kwargs):
        """This method defines a processing pipeline consisting of opening the file
        using the `open` method, applying each of the `proc_funcs` to the output of the previous
        and `save` the processed data using save method."""
        tstr = time.strftime('%Y%m%d')
        files = self.list_files(time)
        files = group_files(files)
        tstr = time.strftime('%Y%m%d')
        filename = f'{self.paths.dst}/{self.name}{self.region.name}_{tstr}.nc'
        if not Path(filename).is_file() or replace:
            data_dict = {v:[] for v in self.bands}
            for file_group in files:
                data = self.open_netcdf4(file_group)
                data = self.resample(data)
                for v in self.bands:
                    data_dict[v].append(data[v])
            kwargs = {'cls': self, **proc_funcs_kwargs}
            for f in proc_funcs:
                data_dict = f(data_dict, time, **kwargs)
            if save:
                self.save(time, data_dict)
            else: return data_dict
        else: warn(f'Skipping files for {time}. File already exists.')

class Viirs375Dataset(ViirsDataset):
    _name = 'VIIRS375'

class Viirs750Dataset(ViirsDataset):
    _name = 'VIIRS750'

# Cell
class MCD64Dataset(BaseDataset):
    "Subclass of `BaseDataset` to process MCD64A1 Collection 6 burned areas product."
    def __init__(self, paths:InOutPath, region:Region, times:pd.DatetimeIndex=None,
                 bands=['bafrac']):
        super().__init__('MCD64A1C6', paths, region, times, bands)

    def list_files(self, time:pd.Timestamp) -> list:
        out = []
        if time in self.times:
            time = pd.Timestamp(f'{time.year}-{time.month}-01')
            time_pattern = f'.A{time.year}{time.dayofyear:03d}.'
            files = self.paths.src.ls(recursive=True, include=['burndate.tif', time_pattern],
                                exclude=['.xml'])
            # Find windows joint with region bounding box
            for f in files:
                data = open_tif(f)
                if not disjoint_bounds(data.bounds, self.region.bbox):
                    out.append(f)
        return out

    def find_dates(self, first:pd.Timestamp=None, last:pd.Timestamp=None):
        pattern = r'^\w+.A(20[0-9][0-9])([0-3][0-9][0-9])..*$'
        times = []
        for f in self.paths.src.ls(recursive=True):
            x = re.search(pattern, f.stem)
            if x is not None:
                year, doy = map(x.group, [1,2])
                times.append(pd.Timestamp(f'{year}-01-01') + pd.Timedelta(days=int(doy)-1))
        self.times = pd.DatetimeIndex(sorted(set(times)))
        if first is not None:
            self.times = self.times[self.times>=first]
        if last is not None:
            self.times = self.times[self.times<=last]
        return self.times

    def file_time_range(self, file) -> pd.DatetimeIndex:
        pattern = r'^\w+.A(20[0-9][0-9])([0-3][0-9][0-9])..*$'
        x = re.search(pattern, file.stem)
        year, doy = map(x.group, [1,2])
        t0 = pd.Timestamp(f'{year}-01-01') + pd.Timedelta(days=int(doy)-1)
        return pd.date_range(t0, periods=monthlen(t0.year, t0.month), freq='D')

    def open(self, files:list) -> dict:
        times = self.file_time_range(files[0])
        data_dict = {'times': times}
        out = np.zeros((len(times), *self.region.shape))
        data = [open_tif(f) for f in files]
        data, tfm = crop(data, self.region.bbox)
        for i, time in enumerate(times):
            x = (data == time.dayofyear).astype(np.int8)
            out[i] += downsample(x, tfm, self.region.transform, self.region.shape)
        data_dict[self.bands[0]] = out
        return data_dict

    def save(self, time:pd.Timestamp, data:dict):
        v = self.bands[0]
        for i, t in enumerate(data['times']):
            super().save(t, {v: data[v][i]})

# Cell
class FireCCI51Dataset(BaseDataset):
    "Subclass of `BaseDataset` to process FireCCI51 burned areas product."
    def __init__(self, paths:InOutPath, region:Region, times:pd.DatetimeIndex=None,
                 bands=['bafrac']):
        super().__init__('FireCCI51', paths, region, times, bands)

    def list_files(self, time:pd.Timestamp) -> list:
        out = []
        if time in self.times:
            time = pd.Timestamp(f'{time.year}-{time.month}-01')
            time_pattern = time.strftime('%Y%m%d')
            files = self.paths.src.ls(recursive=True, include=['JD.tif', time_pattern],
                                exclude=['.xml'])
            # Find windows joint with region bounding box
            for f in files:
                data = open_tif(f)
                if not disjoint_bounds(data.bounds, self.region.bbox):
                    out.append(f)
        return out

    def find_dates(self, first:pd.Timestamp=None, last:pd.Timestamp=None):
        files = self.paths.src.ls(recursive=True, include=['JD.tif'], exclude=['.xml'])
        self.times = pd.DatetimeIndex(sorted(set([pd.Timestamp(o.name[:8]) for o in files])))
        if first is not None:
            self.times = self.times[self.times>=first]
        if last is not None:
            self.times = self.times[self.times<=last]
        return self.times

    def file_time_range(self, file) -> pd.DatetimeIndex:
        t0 = pd.Timestamp(file.name[:8])
        return pd.date_range(t0, periods=monthlen(t0.year, t0.month), freq='D')

    def open(self, files:list) -> dict:
        times = self.file_time_range(files[0])
        data_dict = {'times': times}
        out = np.zeros((len(times), *self.region.shape))
        data = [open_tif(f) for f in files]
        data, tfm = crop(data, self.region.bbox)
        for i, time in enumerate(times):
            x = (data == time.dayofyear).astype(np.int8)
            out[i] += downsample(x, tfm, self.region.transform, self.region.shape)
        data_dict[self.bands[0]] = out
        return data_dict

    def save(self, time:pd.Timestamp, data:dict, do_compression=True):
        v = self.bands[0]
        for i, t in enumerate(data['times']):
            super().save(t, {v: data[v][i]}, do_compression=do_compression)

# Cell
class AusCoverDataset(BaseDataset):
    "Subclass of `BaseDataset` to process AusCover burned areas product."
    def __init__(self, paths:InOutPath, region:Region, times:pd.DatetimeIndex=None,
                bands=['bafrac']):
        super().__init__('AusCover', paths, region, times, bands)

    def list_files(self, time:pd.Timestamp) -> list:
        out = []
        if time.year in self.times.year:
            time = pd.Timestamp(f'{time.year}-01-01')
            time_pattern = time.strftime('_%Y_')
            files = self.paths.src.ls(recursive=True, include=['.tif', time_pattern],
                                exclude=['.xml'])
        return files

    def find_dates(self, first:pd.Timestamp=None, last:pd.Timestamp=None):
        files = self.paths.src.ls(recursive=True, include=['.tif'], exclude=['.xml'])
        self.times = pd.DatetimeIndex(sorted(set([pd.Timestamp(f'{o.stem[-10:-6]}-01-01')
                                                  for o in files])))
        if first is not None:
            self.times = self.times[self.times>=first]
        if last is not None:
            self.times = self.times[self.times<=last]
        return self.times

    def file_time_range(self, file) -> pd.DatetimeIndex:
        t0 = pd.Timestamp(f'{file.stem[-10:-6]}-01-01')
        return pd.date_range(t0, periods=12, freq='MS')

    def open(self, files:list) -> dict:
        times = self.file_time_range(files[0])
        data_dict = {'times': times}
        out = np.zeros((len(times), *self.region.shape))
        data = [open_tif(f) for f in files]
        data = data[0]
        crs = data.crs
        tfm = data.transform
        data = data.read(1)
        for i, time in enumerate(times):
            x = (data == time.month).astype(np.int8)
            out[i] += downsample(x, tfm, self.region.transform,
                                 self.region.shape, src_crs=crs)
        data_dict[self.bands[0]] = out
        return data_dict

    def save(self, time:pd.Timestamp, data:dict, do_compression=True):
        v = self.bands[0]
        for i, t in enumerate(data['times']):
            super().save(t, {v: data[v][i]}, do_compression=do_compression)

# Cell
class MTBSDataset(BaseDataset):
    "Subclass of `BaseDataset` to process MTBS burned areas product."
    def __init__(self, paths:InOutPath, region:Region, times:pd.DatetimeIndex=None,
                bands=['bafrac']):
        super().__init__('MTBS', paths, region, times, bands)

    def list_files(self, *args) -> list:
        files = self.paths.src.ls(recursive=True, include=['.shp'], exclude=['.xml'])
        return files

    def find_dates(self, first:pd.Timestamp=None, last:pd.Timestamp=None):
        files = self.list_files()
        df = open_shp(files[0])
        self.times = pd.date_range(f'{df.Year.min()}-01-01',
                                   f'{df.Year.max()}-12-01', freq='MS')
        if first is not None:
            self.times = self.times[self.times>=first]
        if last is not None:
            self.times = self.times[self.times<=last]
        return self.times

    def open(self, files:list) -> dict:
        data_dict = {'times': self.times}
        data = open_shp(files[0]).to_crs({'init': 'EPSG:4326'})
        out = np.zeros((len(self.times), *self.region.shape))
        R = Region(self.region.name, self.region.bbox, pixel_size=0.0003)
        for i, time in enumerate(self.times):
            x = data.loc[(data.Year==time.year) & (data.StartMonth==time.month)].copy()
            x_raster = rasterize(x, region=R)
            out[i] += downsample(x_raster, R.transform, self.region.transform,
                                self.region.shape)
        data_dict[self.bands[0]] = out
        return data_dict

    def save(self, time:pd.Timestamp, data:dict, do_compression=True):
        v = self.bands[0]
        for i, t in enumerate(data['times']):
            super().save(t, {v: data[v][i]}, do_compression=do_compression)

    def process_all(self, *args):
        self.process_one(self.times[0])

# Cell
class ICNFDataset(BaseDataset):
    "Subclass of `BaseDataset` to process ICNF burned areas product."
    def __init__(self, paths:InOutPath, region:Region, times:pd.DatetimeIndex=None,
                bands=['bafrac']):
        super().__init__('ICNF', paths, region, times, bands)

    def list_files(self, *args) -> list:
        files = self.paths.src.ls(recursive=True, include=['.shp'], exclude=['.xml'])
        return files

    def find_dates(self, first:pd.Timestamp=None, last:pd.Timestamp=None):
        files = self.list_files()
        df = open_shp(files[0])
        self.times = sorted(set([pd.Timestamp(f'{o[:-2]}01')
                            for o in df.FIREDATE if o is not None]))
        if first is not None:
            self.times = self.times[self.times>=first]
        if last is not None:
            self.times = self.times[self.times<=last]
        return self.times

    def open(self, files:list) -> dict:
        data_dict = {'times': self.times}
        data = open_shp(files[0]).to_crs({'init': 'EPSG:4326'})
        data = data.loc[~data.FIREDATE.isna()]
        times = pd.DatetimeIndex([pd.Timestamp(o) for o in data.FIREDATE])
        data['times'] = times
        out = np.zeros((len(self.times), *self.region.shape))
        R = Region(self.region.name, self.region.bbox, pixel_size=0.0003)
        for i, time in enumerate(self.times):
            x = data.loc[(times.year==time.year) &
                         (times.month==time.month)].copy()
            x_raster = rasterize(x, region=R)
            out[i] += downsample(x_raster, R.transform, self.region.transform,
                                self.region.shape)
        data_dict[self.bands[0]] = out
        return data_dict

    def save(self, time:pd.Timestamp, data:dict, do_compression=True):
        v = self.bands[0]
        for i, t in enumerate(data['times']):
            super().save(t, {v: data[v][i]}, do_compression=do_compression)

    def process_all(self, *args):
        self.process_one(self.times[0])

# Cell
class Region2Tiles():
    """Crops processed dataset into 128x128 tiles to be used to train the model"""
    def __init__(self, paths:InOutPath, input_name:str, target_name:str,
                 regions:list=None, bands:list=None, size=128, step=100):
        self.paths = paths
        self.input_name = input_name
        self.target_name = target_name
        self.bands = bands
        self.size = size
        self.step = step
        if regions is None:
            self.regions = [o.name for o in self.paths.src.ls()]
        else:
            self.regions = regions

        for folder in ['images', 'masks']:
            (self.paths.dst/folder).mkdir(exist_ok=True)

    def open(self, file, bands:list):
        "Open .mat file and select `bands`."
        f = sio.loadmat(file)
        return np.array([f[k] for k in bands]).transpose(1,2,0)

    def process_one(self, file, bands, folder):
        """Create tiles for a `file` saving the results in `folder` for each
         crop using `save` method"""
        try:
            data = self.open(file, bands)
            rr, cc, _ = data.shape
            for c in range(0, cc-1, self.step):
                for r in range(0, rr-1, self.step):
                    img = self.crop(data, r, c)
                    if np.nansum(~np.isnan(img)) > 0:
                        self.save(img, file, r, c, folder, bands)
        except:
            warn(f'Unable to process {file}.')

    def process_all(self, max_workers=8, include=[]):
        "Run `process_one` in parallel using the number of workers given by `max_workers`"
        for r in self.regions:
            print(f'Creating tiles for {r}')
            for i, s in enumerate([self.input_name, self.target_name]):
                files_list = self.paths.src.ls(recursive=True, include=[*include, *['.mat', r, s]])
                folder = 'images' if s == self.input_name else 'masks'
                bands = self.bands[i]
                process_one = partial(self.process_one, bands=bands, folder=folder, )
                with ThreadPoolExecutor(max_workers) as e:
                    list(tqdm(e.map(process_one, files_list), total=len(files_list)))

    def crop(self, im, r, c):
        "crop image into a square of size sz"
        sz = self.size
        out_sz = (sz, sz, im.shape[-1])
        rs,cs,hs = im.shape
        tile = np.zeros(out_sz)
        if (r+sz > rs) and (c+sz > cs):
            tile[:rs-r, :cs-c, :] = im[r:, c:, :]
        elif (r+sz > rs):
            tile[:rs-r, :, :] = im[r:, c:c+sz, :]
        elif (c+sz > cs):
            tile[:, :cs-c, :] = im[r:r+sz ,c:, :]
        else:
            tile[...] = im[r:r+sz, c:c+sz, :]
        return tile

    def save(self, data, file, r, c, folder, bands):
        """Save `data` to `file` on `folder` selecting only the given `bands`.
           The row and column index are included in the filename so that the large
           scene can be reconstructed afterwards after generating the model predictions."""
        sio.savemat(self.paths.dst/f'{folder}/{file.stem}_{r}_{c}.mat',
            {v: data[...,i] for i, v in enumerate(bands)}, do_compression=True)


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"filter_files": "00_core.ipynb",
         "ls": "00_core.ipynb",
         "hdf_attr_check": "00_core.ipynb",
         "dict2json": "00_core.ipynb",
         "Path.ls": "06_cli.ipynb",
         "monthlen": "00_core.ipynb",
         "InOutPath": "00_core.ipynb",
         "ProjectPath": "00_core.ipynb",
         "Region": "01_geo.ipynb",
         "open_shp": "01_geo.ipynb",
         "open_tif": "01_geo.ipynb",
         "bounds_from_shapefile": "01_geo.ipynb",
         "size_from_bounds": "01_geo.ipynb",
         "size_resolution_assert": "01_geo.ipynb",
         "rasterize": "01_geo.ipynb",
         "downsample": "01_geo.ipynb",
         "is_intersection": "01_geo.ipynb",
         "polygon_from_bounds": "01_geo.ipynb",
         "crop": "04_predict.ipynb",
         "bounds_from_coords": "01_geo.ipynb",
         "VIIRS375_download": "02_data.ipynb",
         "VIIRS750_download": "02_data.ipynb",
         "BandsFilter": "02_data.ipynb",
         "BandsRename": "02_data.ipynb",
         "MergeTiles": "02_data.ipynb",
         "BandsAssertShape": "02_data.ipynb",
         "ActiveFiresLog": "02_data.ipynb",
         "InterpolateAngles": "02_data.ipynb",
         "MirCalc": "02_data.ipynb",
         "BaseDataset": "02_data.ipynb",
         "mask2nan": "02_data.ipynb",
         "nan2zero": "02_data.ipynb",
         "group_files": "02_data.ipynb",
         "ViirsDataset": "02_data.ipynb",
         "Viirs375Dataset": "02_data.ipynb",
         "Viirs750Dataset": "02_data.ipynb",
         "MCD64Dataset": "02_data.ipynb",
         "FireCCI51Dataset": "02_data.ipynb",
         "AusCoverDataset": "02_data.ipynb",
         "MTBSDataset": "02_data.ipynb",
         "ICNFDataset": "02_data.ipynb",
         "Region2Tiles": "02_data.ipynb",
         "DROP": "03_models.ipynb",
         "BTNorm2d": "03_models.ipynb",
         "LSTM": "03_models.ipynb",
         "SpaceConv": "03_models.ipynb",
         "UpSpaceConv": "03_models.ipynb",
         "TimeConv": "03_models.ipynb",
         "UpTimeConv": "03_models.ipynb",
         "SpaceTimeConv": "03_models.ipynb",
         "UpSpaceTimeConv": "03_models.ipynb",
         "BA_Net": "03_models.ipynb",
         "open_nc": "04_predict.ipynb",
         "image2tiles": "04_predict.ipynb",
         "tiles2image": "04_predict.ipynb",
         "get_preds": "04_predict.ipynb",
         "predict_one": "04_predict.ipynb",
         "save_netcdf": "04_predict.ipynb",
         "inspect_netcdf": "04_predict.ipynb",
         "load_netcdf": "04_predict.ipynb",
         "predict_time": "04_predict.ipynb",
         "predict_month": "04_predict.ipynb",
         "predict_nrt": "04_predict.ipynb",
         "split_mask": "04_predict.ipynb",
         "RunManager": "04c_historical.ipynb",
         "banet_viirs375_download": "06_cli.ipynb",
         "banet_viirs750_download": "06_cli.ipynb",
         "banet_create_dataset": "06_cli.ipynb",
         "banet_dataset2tiles": "06_cli.ipynb",
         "banet_predict_monthly": "06_cli.ipynb",
         "banet_predict_times": "06_cli.ipynb",
         "banet_nrt_run": "06_cli.ipynb",
         "array2png": "07_web.ipynb"}

modules = ["core.py",
           "geo.py",
           "data.py",
           "models.py",
           "predict.py",
           "nrt.py",
           "historical.py",
           "cli.py",
           "web.py"]

doc_url = "https://mnpinto.github.io/banet/"

git_url = "https://github.com/mnpinto/banet/tree/master/"

def custom_doc_links(name): return None
